from unittest.mock import MagicMock, patch

from hiccough import html, html_splat
import pytest
import hiccough
import sys
import copy

from hiccough import build_body_and_attr


def test_simple_div():
    assert html(["div"]) == "<div></div>"


def test_class_shorthand():
    assert html(["div.class"]) == '<div class="class"></div>'


def test_nesting_and_class():
    assert (
        html(
            [
                "div.container>div.section>div.box",
                ["form", ["input", {"type": "checkbox"}]],
            ]
        )
        == '<div class="container"><div class="section"><div class="box"><form><input type="checkbox"></input></form></div></div></div>'
    )


def test_multi_class():
    assert html(["div.a.b"]) == '<div class="a b"></div>'


def test_class_merging():
    assert (
        html(["div.class-a", {"class": "class-b"}])
        == '<div class="class-b class-a"></div>'
    )


def test_id_shorthand():
    assert html(["div#name"]) == '<div id="name"></div>'


def test_multi_id():
    assert html(["div#name#blah"]) == '<div id="name"></div>'


# def test_main():
#     hic = copy.copy(hiccough)
#     hic.html = MagicMock(return_value="")
#     testargs = ["hiccough", '["div"]']
#     with patch.object(sys, "argv", testargs):
#         assert hic.main() == 0
#         hic.html.assert_called_with(["div"])


# def test_main_missing_arg():
#     hiccough.html = MagicMock(return_value="")
#     testargs = ["hiccough"]
#     with patch.object(sys, "argv", testargs):
#         assert hiccough.main() == 1
#         hiccough.html.assert_not_called()


def test_multiple_children():
    assert (
        html(["html", ["head", ["script", {"rel": "stylesheet"}]], ["body>p", "hello"]])
        == '<html><head><script rel="stylesheet"></script></head><body><p>hello</p></body></html>'
    )


def test_multiple_attr():
    assert (
        html(["div", {"id": "not-id", "other": "a"}, {"id": "id", "another": "b"}])
        == '<div id="id" other="a" another="b"></div>'
    )


def test_attr_attaches_to_nested_child():
    assert (
        html(["div>div", {"style": "color: red;"}])
        == '<div><div style="color: red;"></div></div>'
    )


def test_build_body_and_attr():
    assert build_body_and_attr(["test", {"type": "button"}, "thing"]) == (
        "testthing",
        {"type": "button"},
    )


def test_simple_attribute():
    assert (
        html(["div", "", {"style": "color: red;"}]) == '<div style="color: red;"></div>'
    )


def test_simple_nest():
    assert html(["body>p", "hello"]) == "<body><p>hello</p></body>"


def test_new_bug():
    assert (
        html(["div>div>div#target.button", "hello"])
        == '<div><div><div class="button" id="target">hello</div></div></div>'
    )


def test_html_splat():
    assert (
        html_splat("div.thing", "hello", {"id": "id"})
        == '<div id="id" class="thing">hello</div>'
    )



from typing import Any, Dict, List, Tuple, Union
from enum import Enum
import sys
import logging


logger = logging.getLogger("hiccough")
logger.addHandler(logging.NullHandler())

Tag = str
Id = str
Classes = List[str]
Attributes = Dict


class ParseError(Exception):
    pass


def html(val):
    """
    Accepts a list based tree of html nodes of the format
    [tag : str, body: Union[str, list], dict]
    where the tag is the html tag, the body is either text content
    or more nested html and dict is the tag attributes
    """
    logger.info(f"Invoked with {val}")
    # Return string values as is
    if isinstance(val, str):
        return val

    tag, *rest = val
    body, attr = build_body_and_attr(rest)

    attributes = ""

    ## handle custom tag stuff
    tag, rest_tag, tag_attr = tag_handler(tag, attr)

    if rest_tag:
        logger.info(f"Rest Args{[rest_tag, body, attr]}")
        body = html([rest_tag, body, attr])
        attr = {}
        logger.info(f"Rest Body{body}")

    if "class" in attr and "class" in tag_attr:
        attr["class"] += " " + tag_attr["class"]
        del tag_attr["class"]
    elif "class" in tag_attr:
        attr["class"] = tag_attr["class"]
        del tag_attr["class"]
    attr.update(tag_attr)
    logger.info(f"Attr {attr}")
    logger.info(f"TagAttr {tag_attr}")
    for key, val in attr.items():
        attributes += f' {key}="{val}"'
    logger.info(f"Tag : {tag}")
    logger.info(f"Attributes : {attributes}")
    logger.info(f"Body : {body}")
    return f"<{tag}{attributes}>{body}</{tag}>"


RestTag = str


def tag_handler(tag: str, attr: Attributes) -> Tuple[Tag, RestTag, Attributes]:
    # Get first actual tag
    tag, *rest_tags = tag.split(">")
    tag, id, classes = parse_tag(tag)
    temp_attr = {}
    if classes:
        temp_attr["class"] = " ".join(classes)
    if id:
        temp_attr["id"] = id
    rest_tags_txt = ">".join(rest_tags)
    res = (tag, rest_tags_txt, temp_attr)
    logger.info(f"Tag response {res}")
    return res


class TagParserState(Enum):
    TAG = 1
    CLASS = 2
    ID = 3


def parse_tag(tag_str: str) -> Tuple[Tag, Id, Classes]:
    """
    Parses an individual tag of the shorthand div.class-a.class-b#id
    and returns
    """
    if ">" in tag_str:
        raise ParseError(f"Receive nested tag {tag_str}")
    tag: Tag = None
    classes: Classes = []
    id: Id = None

    buffer = ""
    current_state = TagParserState.TAG
    for char in tag_str:
        if char == ".":
            if current_state == TagParserState.TAG:
                tag = buffer

            if current_state == TagParserState.CLASS:
                classes.append(buffer)

            if current_state == TagParserState.ID and not id:
                id = buffer
            buffer = ""
            current_state = TagParserState.CLASS
        elif char == "#":
            if current_state == TagParserState.TAG:
                tag = buffer

            if current_state == TagParserState.CLASS:
                classes.append(buffer)

            if current_state == TagParserState.ID and not id:
                id = buffer
            buffer = ""
            current_state = TagParserState.ID
        else:
            buffer += char

    # empty the buffer
    if current_state == TagParserState.TAG:
        tag = buffer

    if current_state == TagParserState.CLASS:
        classes.append(buffer)

    if current_state == TagParserState.ID and not id:
        id = buffer
    return (tag, id, classes)


def build_body_and_attr(items: List[Any]) -> Tuple[str, Dict]:
    body = ""
    attr = {}
    for item in items:
        # Item is a body element
        if isinstance(item, list):
            body += html(item)
        elif isinstance(item, str):
            body += item
        # Item is an attribute dict
        elif isinstance(item, dict):
            attr.update(item)
        else:
            raise Exception(f"Invalid form exception at: {item} Type: {type(item)}")
    return (body, attr)


def pad_list_none(list: List, pad_size: int):
    """
    Pads a list to the required size with None
    """
    return list + [None] * (pad_size - len(list))


def html_splat(*args):
    """
    A convenience function for Greg to avoid typing an extra [ and ]. Takes
    an arg list and splats it into html()
    """
    return html(args)


def main() -> int:
    """Echo the input arguments to standard output"""
    if len(sys.argv) == 1:
        print("Missing second arg for generating html", file=sys.stderr)
        return 1

    arg = eval(sys.argv[1])
    print(html(arg))
    return 0


if __name__ == "__main__":
    sys.exit(main())
